@@ -24,16 +24,17 @@ import datetime
 import importlib.resources as pkg_resources
 import json
 import logging
-import networkx as nx
 import os
-import pandas as pd
 import re
 import sys
 import textwrap
 import traceback
-
+import typing
 from contextlib import contextmanager
 from io import StringIO
+
+import networkx as nx
+import pandas as pd
 from owlready2 import destroy_entity, get_ontology, onto_path, types,\
                       sync_reasoner_hermit, sync_reasoner_pellet, Thing, Nothing,\
                       AllDisjoint, AllDifferent, DataProperty, ObjectProperty,\
@@ -78,19 +79,19 @@ def load_json(json_file: str) -> dict:
     return data
 
 
-def cleanup(all: bool, *extensions: str) -> None:
+def cleanup(complete: bool, *extensions: str) -> None:
     """ delete all files in the current directory with the extensions specified
 
     :param extensions: extensions of files to be deleted
-    :param all: do not delete current log file if set to False
+    :param complete: do not delete current log file if set to False
     """
-    dir = "./"
+    this_dir = "./"
     for e in extensions:
-        files = [f for f in os.listdir(dir) if f.endswith("." + e)]
-        if not all and LOGFILE in files:
+        files = [f for f in os.listdir(this_dir) if f.endswith("." + e)]
+        if not complete and LOGFILE in files:
             files.remove(LOGFILE)
         for f in files:
-            os.remove(os.path.join(dir, f))
+            os.remove(os.path.join(this_dir, f))
 
 
 class OntoEditor:
@@ -133,7 +134,7 @@ class OntoEditor:
             self.logger.info("ontology file did not exist - created a new one")
 
     @contextmanager
-    def _redirect_to_log(self):
+    def _redirect_to_log(self) -> typing.Iterator[None]:
         with open(os.devnull, "w") as devnull:
             old_stdout = sys.stdout
             old_stderr = sys.stderr
@@ -224,7 +225,7 @@ class OntoEditor:
         """
         with self.onto:
             graph = self.onto_world.as_rdflib_graph()
-            return list(graph.query(query))
+        return list(graph.query(query))
 
     def get_axioms(self) -> list:
         """ identify all axioms included in the onto
@@ -237,6 +238,32 @@ class OntoEditor:
             axioms.append(self.query_onto(self._build_query(query_ax)))
         return axioms
 
+    def add_taxo(self, class_tuples: list) -> None:
+        """ add taxonomy to onto
+
+        :param class_tuples: list of 2-tuples of the form [class, superclass]
+        """
+        with self.onto:
+            for clst in class_tuples:
+                if clst[0] and not clst[1]:
+                    my_class = types.new_class(clst[0], (Thing, ))
+                elif clst[0] and clst[1]:
+                    my_class = types.new_class(clst[0], (self.onto[clst[1]], ))
+                else:
+                    self.logger.warning(f"no class defined: {clst}")
+        self.onto.save(file = self.path)
+
+    @staticmethod
+    def class_dict_to_tuple_list(cls_dict: dict) -> list:
+        """ helper function to convert dict with class definitions to list of
+        tuples as required by add_taxo function
+
+        :param cls_dict: dictionary for taxonomy definition of the form
+            {superclass: [subclasses]}
+        :return: list of class definition 2-tuples of the form [[subclass, superclass], ...]
+        """
+        return [[subcls, supercls] for supercls in cls_dict.keys() for subcls in cls_dict[supercls]]
+
     def add_axioms(self, axiom_tuples: list) -> None:
         """ add entire axioms to onto
         NOTE: only one axiom may be specified at once
@@ -250,18 +277,12 @@ class OntoEditor:
         """
         with self.onto:
             for axiom in axiom_tuples:
-                if axiom[0] and axiom[1] and not axiom[-1]:
-                    my_class = types.new_class(axiom[0], (self.onto[axiom[1]], ))
-                elif axiom[0] and axiom[1] and axiom[-1]:
-                    my_class = types.new_class(axiom[0], (Thing, ))
+                my_class = self.onto[axiom[0]]
+                if axiom[0] and axiom[1] and axiom[-1]:
                     my_class.equivalent_to.append(self.onto[axiom[1]])
-                elif axiom[0] and not axiom[1]:
-                    my_class = types.new_class(axiom[0], (Thing, ))
-                else:
-                    self.logger.warning(f"no class defined: {axiom}")
                 if not axiom[2] and not axiom[4] and not axiom[5] and not axiom[5] == 0 and not axiom[6]:
                     continue
-                if all([axiom[i] for i in [2,4,6]]) or all([axiom[i] for i in [2,4,7]]):
+                if all(axiom[i] for i in [2,4,6]) or all(axiom[i] for i in [2,4,7]):
                     if axiom[-1]:
                         current_axioms = my_class.equivalent_to
                     else:
@@ -300,8 +321,8 @@ class OntoEditor:
                 return
             if resinfo[2] in ["exactly", "max", "min"]:
                 # NOTE: this may be resolved in future versions of Owlready2
-                self.logger.warning(f"qualified cardinality restrictions currently not "
-                                "supported for DPs: {axiom}")
+                self.logger.warning("qualified cardinality restrictions currently not "
+                                    f"supported for DPs: {axiom}")
                 return
         else:
             self.logger.warning(f"restriction includes both op and dp: {axiom}")
@@ -319,7 +340,7 @@ class OntoEditor:
             res = Not(res)
         current_axioms.append(res)
 
-    def _dp_constraint(self, dpres: list):
+    def _dp_constraint(self, dpres: list) -> typing.Optional[ConstrainedDatatype]:
         """
         :param dpres: DP restriction is list of the form [dprange, minex, minin,
             exact, maxin, maxex]
@@ -374,10 +395,10 @@ class OntoEditor:
         :return: True iff expected indices contain values
         """
         indices = [x for x, _ in enumerate(values)]
-        assert all([x in indices for x in expected_values]), "invalid expected_values"
-        test = all([values[i] for i in expected_values]) and\
-               not any([values[i] for i in [e for e in indices if not e in expected_values]])
-        return bool(test)
+        assert all(x in indices for x in expected_values), "invalid expected_values"
+        test = all(values[i] for i in expected_values) and\
+               not any(values[i] for i in [e for e in indices if not e in expected_values])
+        return test
 
     def add_ops(self, op_tuples: list) -> None:
         """ add object properties including their axioms to onto
@@ -518,15 +539,47 @@ class OntoEditor:
         :param label: label to be appended
         :param lang: label's language (optional)
         """
+        entity = self._get_entity_by_name(name)
+        if not entity:
+            return
+        desc = entity.label
+        self._add_description_generic(desc, label, lang)
+
+    def add_annotation(self, name: str, comment: str, lang: str=None) -> None:
+        """ add annotation in language specified as localized string, defaults to
+        regular string if no language is specified
+
+        :param name: entity name
+        :param comment: annotation to append
+        :param lang: annotation's language (optional)
+        """
+        entity = self._get_entity_by_name(name)
+        if not entity:
+            return
+        desc = entity.comment
+        self._add_description_generic(desc, comment, lang)
+
+    def _get_entity_by_name(self, name: str) -> typing.Optional[Thing]:
+        entity = None
         try:
             entity = self.onto[name]
         except AttributeError:
-            self.logger.info(f"unexpected entity: {name}, continuing anyways")
-            return
+            self.logger.info(f"unexpected entity: {name}, return None and continue anyways")
+        return entity
+
+    def _add_description_generic(self, desc_list: list, description: str,
+                                 lang: typing.Optional[str]) -> None:
+        """ add description in language specified as localized string, defaults to
+        regular string if no language is specified
+
+        :param desc_list: list to which to append the description
+        :param description: description to append
+        :param lang: description's language (optional)
+        """
         if lang:
-            entity.label.append(locstr(label, lang=lang))
+            desc_list.append(locstr(description, lang=lang))
         else:
-            entity.label.append(label)
+            desc_list.append(description)
         self.onto.save(file = self.path)
 
     def remove_from_taxo(self, elem_list: list, reassign: bool=True) -> None:
@@ -547,8 +600,8 @@ class OntoEditor:
                 descendants.remove(self.onto[elem])
                 individuals = list(self.onto[elem].instances())
                 if reassign:
-                    sc_res = self.get_class_restrictions(self.onto[elem].name, "is_a")
-                    eq_res = self.get_class_restrictions(self.onto[elem].name, "equivalent_to")
+                    sc_res = self.get_class_restrictions(self.onto[elem].name, res_type="is_a")
+                    eq_res = self.get_class_restrictions(self.onto[elem].name, res_type="equivalent_to")
                 for desc in descendants:
                     desc.is_a.append(parent[0])
                     if reassign:
@@ -558,7 +611,7 @@ class OntoEditor:
                 destroy_entity(self.onto[elem])
         self.onto.save(file = self.path)
 
-    def get_class_restrictions(self, class_name: str, res_only: bool=True, res_type: str="is_a") -> list:
+    def get_class_restrictions(self, class_name: str, res_type: str="is_a", res_only: bool=True) -> list:
         """ retrieve restrictions on specific class by restriction type
 
         :param class_name: name of the class for which restrictions shall be returned
@@ -577,7 +630,7 @@ class OntoEditor:
                 sys.exit(1)
             if res_only:
                 elems = [x for x in elems if isinstance(x, Restriction)]
-            return elems
+        return elems
 
     def remove_restrictions_on_class(self, class_name: str) -> None:
         """ remove all restrictions on a given class
@@ -638,7 +691,7 @@ class OntoEditor:
                 inconsistent_classes = list(inf_onto.inconsistent_classes())
             except Exception as exc:
                 if reasoner == "pellet" and debug:
-                    inconsistent_classes = self._analyze_pellet_results(exc)
+                    inconsistent_classes = self._analyze_pellet_results(str(exc))
                 else:
                     inconsistent_classes = self.reasoning("pellet", False, True)
         if inconsistent_classes:
@@ -655,12 +708,12 @@ class OntoEditor:
         if reasoner not in reasoners:
             self.logger.warning(f"unexpected reasoner: {reasoner} - available reasoners: {reasoners}")
 
-    def _analyze_pellet_results(self, exc: Exception) -> list:
+    def _analyze_pellet_results(self, exc: str) -> list:
         """ analyze the explanation returned by Pellet, print it and return
         inconsistent classes
         IDEA: also consider restrictions on properties and facts about instances
 
-        :param exc: exception thrown during reasoning process
+        :param exc: string of exception thrown during reasoning process
         :return: list of classes identified as problematic
         """
         inconsistent_classes = []
@@ -683,7 +736,7 @@ class OntoEditor:
         :param pellet_traceback: traceback created when running reasoner
         :return: tuple of entire explanation and list of axioms included in explanation
         """
-        rex = re.compile("Explanation\(s\): \n(.*?)\n\n", re.DOTALL|re.MULTILINE)
+        rex = re.compile(r"Explanation\(s\): \n(.*?)\n\n", re.DOTALL|re.MULTILINE)
         res = set(re.findall(rex, pellet_traceback))
         axioms: list=[]
         if res:
@@ -735,9 +788,9 @@ class OntoEditor:
         :param restype: type of class restriction, either is_a or equivalent_to
         :return: list of class restrictions for inconsistent_classes - does not return parent classes
         """
-        return [self.get_class_restrictions(ic.name, res_only=True, res_type=restype) for ic in inconsistent_classes]
+        return [self.get_class_restrictions(ic.name, res_type=restype, res_only=True) for ic in inconsistent_classes]
 
-    def _interactively_delete_axs_by_rel(self, rel: str, classes: list, axioms: list, msg: str) -> None:
+    def _interactively_delete_axs_by_rel(self, rel: str, classes: list, axioms: dict, msg: str) -> None:
         """
         :param rel: relation between class and axioms - is_a or equivalent_to
         :param classes: classes for which axioms are to be removed
@@ -755,7 +808,7 @@ class OntoEditor:
                     # IDEA: instead of simply deleting axioms, also allow user to edit them
 
     @staticmethod
-    def _bool_user_interaction(question: str, info: str=None) -> str:
+    def _bool_user_interaction(question: str, info: str=None) -> bool:
         """ simple CLI for yes/ no/ quit interaction
         """
         answer = {"y": True,
@@ -767,11 +820,11 @@ class OntoEditor:
         while user_input not in ["y", "n", "q"]:
             print("invalid choice, please try again")
             user_input = input()
-        if user_input in ["y", "n"]:
-            return answer[user_input]
         if user_input == "q":
             print("quitting - process needs to be restarted")
             sys.exit(0)
+        else:
+            return answer[user_input]
 
     @staticmethod
     def _remove_nt_brackets(triple: list) -> list:
@@ -802,10 +855,10 @@ class OntoEditor:
                     n[1]["color"] = color
         return nxgraph
 
-    def _ntriples_to_df(self) -> nx.MultiDiGraph:
+    def _ntriples_to_df(self) -> pd.DataFrame:
         self.export_ntriples()
-        f = open(self.path.rsplit(".", 1)[0] + ".nt", "r")
-        lines = f.readlines()
+        with open(self.path.rsplit(".", 1)[0] + ".nt", "r") as f:
+            lines = f.readlines()
         df = pd.DataFrame(columns=["subject", "predicate", "object"])
         for rownum, row in enumerate(lines):
             df.loc[rownum] = self._remove_nt_brackets(row.rsplit(".", 1)[0].split(" ")[:3])
@@ -813,7 +866,7 @@ class OntoEditor:
 
     @staticmethod
     def _query_results_to_df(query_results: list) -> pd.DataFrame:
-        clean_data = [[str(elem).split("#")[-1] for elem in row] for row in query_results]
+        clean_data = [[str(elem).rsplit("#", maxsplit=1)[-1] for elem in row] for row in query_results]
         df = pd.DataFrame(clean_data, columns=['subject', 'predicate', 'object'])
         return df
 
@@ -853,10 +906,10 @@ class OntoEditor:
         max_radius = 5
         nodes_to_be_ignored = ["owl:Class", "owl:Thing", "owl:NamedIndividual", "owl:Restriction"]
 
-        if show_class_descendants:
+        if classes and show_class_descendants:
             descendent_lists = [[desc.name for desc in self.onto[c].descendants()] for c in classes]
             subclasses = list({c for sublist in descendent_lists for c in sublist})
-        else:
+        elif classes and not show_class_descendants:
             subclasses = classes
 
         def _sparql_set_values(node, values):
